/*
 *  soapUI, copyright (C) 2006 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of the GNU Lesser General Public License as published by the Free Software Foundation; 
 *  either version 2.1 of the License, or (at your option) any later version.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.eviware.soapui.config.TestCaseConfig;
import com.eviware.soapui.config.TestSuiteConfig;
import com.eviware.soapui.config.TestSuiteRunTypesConfig;
import com.eviware.soapui.config.TestSuiteRunTypesConfig.Enum;
import com.eviware.soapui.impl.actions.ShowDesktopPanelAction;
import com.eviware.soapui.impl.wsdl.actions.support.ShowOnlineHelpAction;
import com.eviware.soapui.impl.wsdl.actions.testsuite.AddNewTestCaseAction;
import com.eviware.soapui.impl.wsdl.actions.testsuite.CloneTestSuiteAction;
import com.eviware.soapui.impl.wsdl.actions.testsuite.RemoveTestSuiteAction;
import com.eviware.soapui.impl.wsdl.actions.testsuite.RenameTestSuiteAction;
import com.eviware.soapui.impl.wsdl.loadtest.WsdlLoadTest;
import com.eviware.soapui.impl.wsdl.support.HelpUrls;
import com.eviware.soapui.impl.wsdl.testcase.WsdlTestCase;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestStep;
import com.eviware.soapui.model.testsuite.TestCase;
import com.eviware.soapui.model.testsuite.TestSuite;
import com.eviware.soapui.model.testsuite.TestSuiteListener;
import com.eviware.soapui.support.action.ActionSupport;

/**
 * TestSuite implementation for WSDL projects.
 * 
 * @author Ole.Matzura
 */

public class WsdlTestSuite extends AbstractWsdlModelItem<TestSuiteConfig> implements TestSuite
{
   private final WsdlProject project;
   private List<WsdlTestCase> testCases = new ArrayList<WsdlTestCase>();
   private Set<TestSuiteListener> listeners = new HashSet<TestSuiteListener>();
   
   public WsdlTestSuite(WsdlProject project, TestSuiteConfig config)
   {
   	super( config, project, "/testSuite.gif" );
      this.project = project;
      
      List<TestCaseConfig> testCaseConfigs = config.getTestCaseList();
      for (int i = 0; i < testCaseConfigs.size(); i++)
      {
         testCases.add( new WsdlTestCase( this, testCaseConfigs.get(i) ));
      }
      
      addAction( new ShowDesktopPanelAction( "Open TestSuite Editor", 
      		"Opens the TestSuite Editor for this TestSuite", this ));
      addAction( ActionSupport.SEPARATOR_ACTION );
      addAction( new AddNewTestCaseAction( this ) );
      addAction( new CloneTestSuiteAction( this ));
      addAction( ActionSupport.SEPARATOR_ACTION );
      addAction( new RenameTestSuiteAction( this ) );
      addAction( new RemoveTestSuiteAction( this ) );
      addAction( ActionSupport.SEPARATOR_ACTION );
      addAction( new ShowOnlineHelpAction( HelpUrls.TESTSUITE_HELP_URL ));
      
      if( !config.isSetRunType() )
      	config.setRunType( TestSuiteRunTypesConfig.SEQUENTIAL );
   }
   
	public TestSuiteRunType getRunType()
	{
		Enum runType = getConfig().getRunType();
		
		if( runType.equals( TestSuiteRunTypesConfig.PARALLELL ))
			return TestSuiteRunType.PARALLEL;
		else
			return TestSuiteRunType.SEQUENTIAL;
	}
	
	public void setRunType( TestSuiteRunType runType )
	{
		TestSuiteRunType oldRunType = getRunType();
		
		if( runType == TestSuiteRunType.PARALLEL && oldRunType != TestSuiteRunType.PARALLEL )
		{
			getConfig().setRunType( TestSuiteRunTypesConfig.PARALLELL );
			notifyPropertyChanged( RUNTYPE_PROPERTY, oldRunType, runType );
		}
		else if( runType == TestSuiteRunType.SEQUENTIAL && oldRunType != TestSuiteRunType.SEQUENTIAL )
		{
			getConfig().setRunType( TestSuiteRunTypesConfig.SEQUENTIAL );
			notifyPropertyChanged( RUNTYPE_PROPERTY, oldRunType, runType );
		}
	}

	public WsdlProject getProject()
   {
      return project;
   }

   public int getTestCaseCount()
   {
      return testCases.size();
   }

   public WsdlTestCase getTestCaseAt(int index)
   {
      return testCases.get( index );
   }

   public WsdlTestCase getTestCaseByName(String testCaseName)
	{
		return ( WsdlTestCase ) getWsdlModelItemByName( testCases, testCaseName );
	}

	public WsdlTestCase cloneTestCase( WsdlTestCase testCase, String name )
   {
   	TestCaseConfig newTestCase = getConfig().addNewTestCase();
		newTestCase.set( testCase.getConfig() );
		newTestCase.setName( name );
		WsdlTestCase newWsdlTestCase = new WsdlTestCase( this, newTestCase );
		
      testCases.add( newWsdlTestCase );
      fireTestCaseAdded( newWsdlTestCase );

      return newWsdlTestCase;
   }
   
   public WsdlTestCase addNewTestCase( String name )
   {
      WsdlTestCase testCase = new WsdlTestCase( this, getConfig().addNewTestCase());
      testCase.setName( name );
      testCase.setFailOnError( true );
      testCase.setSearchProperties( true );
      testCases.add( testCase );
      fireTestCaseAdded( testCase );

      return testCase;
   }
   
   public void removeTestCase(WsdlTestCase testCase )
   {
      int ix = testCases.indexOf( testCase );

      testCases.remove( ix );
      try
      {
      	fireTestCaseRemoved( testCase );
      }
      finally
      {
      	testCase.release();
      	getConfig().removeTestCase( ix );
      }
   }

   public void fireTestCaseAdded( WsdlTestCase testCase )
   {
      TestSuiteListener[] a = listeners.toArray( new TestSuiteListener[listeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testCaseAdded( testCase );
      }
   }
   
   public void fireTestCaseRemoved( WsdlTestCase testCase )
   {
   	TestSuiteListener[] a = listeners.toArray( new TestSuiteListener[listeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testCaseRemoved( testCase );
      }
   } 

   public void fireTestStepAdded( WsdlTestStep testStep, int index )
   {
   	TestSuiteListener[] a = listeners.toArray( new TestSuiteListener[listeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testStepAdded( testStep, index );
      }
   }
   
   public void fireTestStepRemoved( WsdlTestStep testStep, int ix )
   {
   	TestSuiteListener[] a = listeners.toArray( new TestSuiteListener[listeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testStepRemoved( testStep, ix );
      }
   } 
   
   public void fireTestStepMoved( WsdlTestStep testStep, int ix, int offset )
   {
   	TestSuiteListener[] a = listeners.toArray( new TestSuiteListener[listeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testStepMoved( testStep, ix, offset );
      }
   } 

   public void fireLoadTestAdded( WsdlLoadTest loadTest )
   {
   	TestSuiteListener[] a = listeners.toArray( new TestSuiteListener[listeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].loadTestAdded( loadTest );
      }
   }
   
   public void fireLoadTestRemoved( WsdlLoadTest loadTest )
   {
   	TestSuiteListener[] a = listeners.toArray( new TestSuiteListener[listeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].loadTestRemoved( loadTest );
      }
   } 
   
	public void addTestSuiteListener(TestSuiteListener listener)
	{
		listeners.add( listener );
	}

	public void removeTestSuiteListener(TestSuiteListener listener)
	{
		listeners.remove( listener );
	}

	public int getTestCaseIndex(TestCase testCase)
	{
		return testCases.indexOf( testCase );
	}

	public void release()
	{
		super.release();
		
		for( WsdlTestCase testCase : testCases )
			testCase.release();
	} 
	
	
}
