/*
 *  soapUI, copyright (C) 2006 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of the GNU Lesser General Public License as published by the Free Software Foundation; 
 *  either version 2.1 of the License, or (at your option) any later version.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.support.http;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.security.GeneralSecurityException;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.zip.GZIPInputStream;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpMethodBase;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.contrib.ssl.EasySSLProtocolSocketFactory;
import org.apache.commons.httpclient.protocol.Protocol;
import org.apache.commons.httpclient.protocol.ProtocolSocketFactory;
import org.apache.commons.ssl.KeyMaterial;
import org.apache.log4j.Logger;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.model.settings.Settings;
import com.eviware.soapui.model.settings.SettingsListener;
import com.eviware.soapui.settings.HttpSettings;
import com.eviware.soapui.settings.SSLSettings;

/**
 * HttpClient related tools
 * 
 * @author Ole.Matzura
 */

public class HttpClientSupport
{
	private final static Helper helper = new Helper();

	/**
	 * Internal helper to ensure synchronized access..
	 */

	private static class Helper
	{
		private HttpClient httpClient;
		private final static Logger log = Logger.getLogger( HttpClientSupport.Helper.class );
		private EasySSLProtocolSocketFactory easySSL;

		public Helper()
		{
			try
			{
				easySSL = new EasySSLProtocolSocketFactory();
				initSSL( easySSL );

				Protocol easyhttps = new Protocol( "https", ( ProtocolSocketFactory ) easySSL, 443 );
				Protocol.registerProtocol( "https", easyhttps );
			}
			catch( Throwable e )
			{
				log.error( "Failed to initialize SSL settings; " + e );
			}

			MultiThreadedHttpConnectionManager connectionManager = new MultiThreadedHttpConnectionManager();
			connectionManager.getParams().setDefaultMaxConnectionsPerHost( 500 );
			connectionManager.getParams().setMaxTotalConnections( 2000 );
			httpClient = new HttpClient( connectionManager );
			
			SoapUI.getSettings().addSettingsListener( new SettingsListener() {

				public void settingChanged( String name, String newValue, String oldValue )
				{
					if( name.equals( SSLSettings.KEYSTORE ) || name.equals( SSLSettings.KEYSTORE_PASSWORD ))
					{
						try
						{
							log.info( "Updating keyStore.." );
							initKeyMaterial( easySSL );
						}
						catch( Throwable e )
						{
							log.error( "Failed to initialize SSL settings; " + e );
						}
					}
				}} );
		}

		private void initSSL( EasySSLProtocolSocketFactory easySSL ) throws IOException,
		GeneralSecurityException
		{
			log.info( "Initializing SSL" );
			initKeyMaterial( easySSL );

			 /*
	       Commented out for now - EasySSLProtocolSocketFactory already
	       trusts everything!  Below is some code that might work for when
	       SoapUI moves away from "EasySSLProtocolSocketFactory".                     

	    String trustStore = settings.getString( SSLSettings.TRUSTSTORE, null );
	    trustStore = trustStore != null ? trustStore.trim() : "";
	    pass = settings.getString( SSLSettings.TRUSTSTORE_PASSWORD, "" );
	    pwd = pass.toCharArray();
	    if ( !"".equals( trustStore ) ) {
	        File f = new File( trustStore );
	        if ( f.exists() ) {
	            TrustMaterial tm = null;
	            try
	            {
	                tm = new TrustMaterial( trustStore, pwd );
	            }
	            catch ( GeneralSecurityException gse )
	            {
	                String trimmedPass = pass.trim();
	                if ( "".equals( trimmedPass ) )
	                {
	                    // If the password is all spaces, then we'll allow
	                    // loading of the TrustMaterial without a password.
	                    tm = new TrustMaterial( trustStore );
	                }
	                else
	                {
	                    log.error( "Failed to load TrustMaterial: " + gse );
	                }
	            }
	            if ( tm != null )
	            {
	                easySSL.setTrustMaterial( tm );
	                log.info( "Added TrustStore from file [" + trustStore + "]" );
	            }
	        } else {
	            log.error( "Missing trustStore [" + trustStore + "]" );
	        }
	    }
*/
		}

		private void initKeyMaterial( EasySSLProtocolSocketFactory easySSL ) throws IOException, NoSuchAlgorithmException, KeyStoreException, KeyManagementException, CertificateException
		{
			Settings settings = SoapUI.getSettings();
			
			String keyStore = settings.getString( SSLSettings.KEYSTORE, null );
			keyStore = keyStore != null ? keyStore.trim() : "";
			String pass = settings.getString( SSLSettings.KEYSTORE_PASSWORD, "" );
			char[] pwd = pass.toCharArray();
			if( !"".equals( keyStore ) )
			{
				File f = new File( keyStore );
				if( f.exists() )
				{
					KeyMaterial km = null;
					try
					{
						km = new KeyMaterial( keyStore, pwd );
						log.info( "Set KeyMaterial from file [" + keyStore + "]" );
					}
					catch( GeneralSecurityException gse )
					{
						log.error( "Failed to load KeyMaterial [" + keyStore + "] " + gse );
					}
					if( km != null )
					{
						easySSL.setKeyMaterial( km );
					}
				}
			}
			else
			{
				easySSL.setKeyMaterial( null );
			}
		}

		public HttpClient getHttpClient()
		{
			return httpClient;
		}
	}

	public static HttpClient getHttpClient()
	{
		return helper.getHttpClient();
	}

	public static void applyHttpSettings( HttpMethodBase postMethod, Settings settings )
	{
		// user agent?
		String userAgent = settings.getString( HttpSettings.USER_AGENT, null );
		if( userAgent != null && userAgent.length() > 0 )
			postMethod.setRequestHeader( "User-Agent", userAgent );

		// timeout?
		long timeout = settings.getLong( HttpSettings.SOCKET_TIMEOUT,
					HttpSettings.DEFAULT_SOCKET_TIMEOUT );
		postMethod.getParams().setSoTimeout( ( int ) timeout );
	}

	public static boolean isZippedResponse( HttpMethodBase method )
	{
		Header contentType = method.getResponseHeader( "Content-Type" );
		Header contentEncoding = method.getResponseHeader( "Content-Encoding" );

		if( contentType != null && contentType.getValue().toUpperCase().endsWith( "GZIP" ) )
			return true;

		if( contentEncoding != null && contentEncoding.getValue().toUpperCase().endsWith( "GZIP" ) )
			return true;

		return false;
	}

	public static byte[] decompress( byte[] s )
	{
		GZIPInputStream zipin;
		try
		{
			InputStream in = new ByteArrayInputStream( s );
			zipin = new GZIPInputStream( in );
		}
		catch( IOException e )
		{
			e.printStackTrace();
			return s;
		}

		int sizeBuffer = 8192;
		byte[] buffer = new byte[sizeBuffer];
		try
		{
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			int length;
			while( ( length = zipin.read( buffer, 0, sizeBuffer ) ) != -1 )
				out.write( buffer, 0, length );
			s = out.toByteArray();
			out.close();
		}
		catch( IOException e )
		{
			e.printStackTrace();
			return s;
		}
		try
		{
			zipin.close();
		}
		catch( IOException e )
		{
		}

		return s;
	}
}
