/*
 *  soapUI, copyright (C) 2006 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of the GNU Lesser General Public License as published by the Free Software Foundation; 
 *  either version 2.1 of the License, or (at your option) any later version.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.support.components;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Toolkit;
import java.awt.datatransfer.StringSelection;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.TransferHandler;
import javax.swing.border.TitledBorder;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableModel;

import org.apache.commons.beanutils.PropertyUtils;

/**
 * Table for displaying property name/value pairs
 * 
 * @author Ole.Matzura
 */

public class JPropertiesTable<T> extends JPanel
{
	public final static Object[] BOOLEAN_OPTIONS = new Object[] {Boolean.TRUE, Boolean.FALSE }; 
	
	private PropertiesTableModel<T> tableModel;
	private JTable table;

	public JPropertiesTable( String title )
	{
		this( title, null );
	}
	
	public JPropertiesTable( String title, T propertyObject )
	{
		super( new BorderLayout());
		
		tableModel = new PropertiesTableModel<T>( propertyObject );
		table = new PTable( tableModel );
		
		table.getColumnModel().getColumn(0).setHeaderValue( "Property" );
      table.getColumnModel().getColumn(1).setHeaderValue( "Value" );
      table.getColumnModel().getColumn(0).setCellRenderer( new PropertiesTableCellRenderer() );
      table.getColumnModel().getColumn(1).setCellRenderer( new PropertiesTableCellRenderer() );
      
      add( new JScrollPane( table ), BorderLayout.CENTER );
      if( title != null )
      {
	      TitledBorder titledBorder = BorderFactory.createTitledBorder( BorderFactory.createEmptyBorder(), title );
	      titledBorder.setTitleFont( titledBorder.getTitleFont().deriveFont( Font.PLAIN, 11 ));
			setBorder( titledBorder);
      }
		
		table.setBackground( Color.WHITE );
		setPreferredSize( table.getPreferredSize());
	}
	
	public void setPropertyObject( T propertyObject )
	{
		if( table.isEditing() )
			table.getCellEditor().stopCellEditing();
		
		tableModel.setPropertyObject( propertyObject );
	}
	
	public PropertiesTableModel getTableModel()
	{
		return tableModel;
	}

	public void addProperty( String caption, String name)
	{
		addProperty( caption, name, false  );
	}
	
	public void addProperty( String caption, String name, boolean editable )
	{
		addProperty( caption, name, editable, null );
	}
	
	public void addProperty( String caption, String name, boolean editable, PropertyFormatter formatter )
	{
		tableModel.addProperty( caption, name, editable, formatter  );
	}
	
	private static final class PropertiesTableModel<T> extends AbstractTableModel implements PropertyChangeListener 
   {
		private List<PropertyDescriptor> properties = new ArrayList<PropertyDescriptor>();
		private T propertyObject;
		
		public PropertiesTableModel(T propertyObject)
		{
			this.propertyObject = propertyObject;
		}
		
		public void setPropertyObject(T propertyObject)
		{
			this.propertyObject = propertyObject;
			fireTableDataChanged();
		}

		public void addProperty( String caption, String name, boolean editable, PropertyFormatter formatter )
		{
			properties.add( new PropertyDescriptor( caption, name, editable, formatter ));
		}
		
		public void addProperty( String caption, String name, Object [] options )
		{
			properties.add( new PropertyDescriptor( caption, name, options ));
		}
		
		public int getRowCount()
		{
			return properties.size();
		}

		public int getColumnCount()
		{
			return 2;
		}
		
		public boolean isCellEditable(int rowIndex, int columnIndex)
		{
			if( columnIndex == 0 || propertyObject == null ) return false;
			return properties.get( rowIndex ).isEditable() && 
				PropertyUtils.isWriteable( propertyObject, properties.get( rowIndex ).getName() );
		}
		
		public void setValueAt(Object aValue, int rowIndex, int columnIndex)
		{
			try
			{
				if( propertyObject != null && columnIndex == 1 && properties.get( rowIndex ).isEditable() )
				{
					PropertyUtils.setSimpleProperty( propertyObject, properties.get( rowIndex ).getName(), aValue );
				}
			}
			catch (IllegalAccessException e)
			{
				e.printStackTrace();
			}
			catch (InvocationTargetException e)
			{
				e.printStackTrace();
			}
			catch (NoSuchMethodException e)
			{
				e.printStackTrace();
			}
		}

		public Object getValueAt(int rowIndex, int columnIndex)
		{
			if( propertyObject == null )
				return null;
			
			try
			{
				PropertyDescriptor propertyDescriptor = properties.get(rowIndex);
				switch (columnIndex)
				{
					case 0:
						return propertyDescriptor.getCaption();
					case 1:
					{
						Object value = PropertyUtils.getSimpleProperty(propertyObject,propertyDescriptor.getName());
						return propertyDescriptor.getFormatter().format( propertyDescriptor.getName(), value );
					}
				}
			}
			catch (IllegalAccessException e)
			{
				e.printStackTrace();
			}
			catch (InvocationTargetException e)
			{
				e.printStackTrace();
			}
			catch (NoSuchMethodException e)
			{
				e.printStackTrace();
			}
			
			return null;
		}
		
		public PropertyDescriptor getPropertyDescriptorAt( int row )
		{
			return properties.get( row );
		}

		public void propertyChange(PropertyChangeEvent evt)
		{
			String name = evt.getPropertyName();
			if( name.lastIndexOf( '@') > 0 ) 
				name = name.substring( name.lastIndexOf( '@') + 1);
			
			for( int c = 0; c < properties.size(); c++ )
			{
				if( properties.get( c ).getName().equals( name ))
				{
					fireTableCellUpdated( c, 1 );
					return;
				}
			}
		}}
	
	private static class PropertyDescriptor
	{
		private final String caption;
		private final String name;
		private boolean editable;
		private PropertyFormatter formatter ;
		private Object [] options;
		private DefaultCellEditor cellEditor;

		public PropertyDescriptor(String caption, String name, boolean editable, PropertyFormatter formatter)
		{
			this.caption = caption;
			this.name = name;
			this.editable = editable;
			this.formatter = formatter;
			
			JTextField textField = new JTextField();
			textField.setBorder(BorderFactory.createEmptyBorder()); 
			cellEditor = new DefaultCellEditor( textField );
		}

		public PropertyDescriptor(String caption, String name, Object[] options)
		{
			this.caption = caption;
			this.name = name;
			this.options = options;
			editable = true;
			
			JComboBox comboBox = new JComboBox( options );
			
			if( options[0] == null )
			{
				comboBox.setEditable( true );
				comboBox.removeItemAt( 0 );
			}
			
			comboBox.setBorder(null); 
			cellEditor = new DefaultCellEditor( comboBox );
		}

		public void setFormatter( PropertyFormatter formatter )
		{
			this.formatter = formatter;
		}
		
		public PropertyFormatter getFormatter()
		{
			return formatter == null ? DefaultFormatter.getInstance() : formatter;
		}

		public String getCaption()
		{
			return caption;
		}

		public boolean isEditable()
		{
			return editable;
		}
		
		public Object[] getOptions()
		{
			return options;
		}
		
		public boolean hasOptions()
		{
			return options != null;
		}

		public String getName()
		{
			return name;
		}
		
		public TableCellEditor getCellEditor()
		{
			return cellEditor;
		}
	}

	private static class PropertiesTableCellRenderer extends DefaultTableCellRenderer
	{
		public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column)
		{
			Component component = super.getTableCellRendererComponent(table, value, isSelected, hasFocus,
					row, column);
			
			if( component instanceof JComponent && value != null )
				((JComponent)component).setToolTipText( value.toString() );
			
			return component;
		}
	}
	/*
	private class PropertiesTableCellEditor extends AbstractCellEditor implements TableCellEditor
	{
		private JTextField textField;
		private JComboBox comboBox;
		private JComponent current;
		
		public PropertiesTableCellEditor()
		{
			textField = new JTextField();
			comboBox = new JComboBox();
			comboBox.putClientProperty("JComboBox.isTableCellEditor", Boolean.TRUE);
		}

		public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row, int column)
		{
			PropertyDescriptor descriptor = tableModel.getPropertyDescriptorAt( row );
			
			if( descriptor.hasOptions())
			{
				comboBox.setModel( new DefaultComboBoxModel( descriptor.getOptions() ));
				comboBox.setSelectedItem( value );
				current = comboBox;
			}
			else
			{
				textField.setText( value == null ? "" : value.toString() );
				current = textField;
			}
			
			current.setBorder( null );
			current.setBackground( Color.WHITE );
			
			return current;
		}

		public Object getCellEditorValue()
		{
			return current == comboBox ? comboBox.getSelectedItem() : textField.getText();
		}
	}
	*/
	
	/**
	 * Formatter used for displaying property values
	 * 
	 * @author Ole.Matzura
	 */
	
	public interface PropertyFormatter
	{
		public Object format( String propertyName, Object value );
	}

	private static class DefaultFormatter implements PropertyFormatter
	{
		private static PropertyFormatter instance;

		public static PropertyFormatter getInstance()
		{
			if( instance == null )
				instance = new DefaultFormatter();
			
			return instance;
		}

		public Object format(String propertyName, Object value)
		{
			return value;
		}
	}

	public void addProperty(String caption, String name, Object [] options)
	{
		tableModel.addProperty( caption, name, options );
	}
	
	private class PTable extends JTable
	{
		public PTable( TableModel tableModel )
		{
			super( tableModel );
			
		//	setAutoStartEditOnKeyStroke( true );
			
			getActionMap().put( TransferHandler.getCopyAction().getValue(Action.NAME), new AbstractAction()  
			{
				public void actionPerformed(ActionEvent e)
				{
					int row = getSelectedRow();
					if( row == -1 )
						return;
					
					StringSelection selection = new StringSelection( getValueAt( row, 1 ).toString() ); 
					Toolkit.getDefaultToolkit().getSystemClipboard().setContents( selection, selection );
				}});
			
			putClientProperty("terminateEditOnFocusLost", Boolean.TRUE);
		/*	
			addFocusListener( new FocusAdapter() {

				public void focusLost(FocusEvent e)
				{
					if( isEditing() && getCellEditor() != null )
						getCellEditor().stopCellEditing();
				}}  );*/
		}

		public TableCellEditor getCellEditor(int row, int column)
		{
			if( column == 0 )
				return super.getCellEditor(row, column);
			else
				return tableModel.getPropertyDescriptorAt( row ).getCellEditor();
		}
	}
}
